# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2018 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six
from six.moves import collections_abc

from restalchemy.dm import filters
from restalchemy.dm import types
from restalchemy.storage.sql.dialect.query_builder import common
from restalchemy.storage.sql import utils

LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class AbstractClause(object):

    def __init__(self, column, value_type, value):
        super(AbstractClause, self).__init__()
        self._value = self._convert_value(value_type, value)
        self._column = column

    def _convert_value(self, value_type, value):
        return value_type.to_simple_type(value)

    @property
    def value(self):
        return self._value

    @property
    def column(self):
        return (self._column.compile()
                if isinstance(self._column, common.ColumnFullPath)
                else self._column)

    @abc.abstractmethod
    def construct_expression(self):
        raise NotImplementedError()


class EQ(AbstractClause):

    def construct_expression(self):
        return ("%s = " % self.column) + "%s"


class NE(AbstractClause):

    def construct_expression(self):
        return ("%s <> " % self.column) + "%s"


class GT(AbstractClause):

    def construct_expression(self):
        return ("%s > " % self.column) + "%s"


class GE(AbstractClause):

    def construct_expression(self):
        return ("%s >= " % self.column) + "%s"


class LT(AbstractClause):

    def construct_expression(self):
        return ("%s < " % self.column) + "%s"


class LE(AbstractClause):

    def construct_expression(self):
        return ("%s <= " % self.column) + "%s"


class Is(AbstractClause):

    def construct_expression(self):
        return ("%s IS " % self.column) + "%s"


class IsNot(AbstractClause):

    def construct_expression(self):
        return ("%s IS NOT " % self.column) + "%s"


class In(AbstractClause):

    def _convert_value(self, value_type, value):
        # Note(efrolov): Replace empty list with [Null]. Some SQL servers
        #                forbid empty lists in "in" operator.
        return [value_type.to_simple_type(item) for item in value] or [None]

    def construct_expression(self):
        return ("%s IN " % self.column) + "%s"


class NotIn(In):

    def construct_expression(self):
        return ("%s NOT IN " % self.column) + "%s"


class Like(AbstractClause):

    def construct_expression(self):
        return ("%s LIKE " % self.column) + "%s"


class NotLike(AbstractClause):

    def construct_expression(self):
        return ("%s NOT LIKE " % self.column) + "%s"


@six.add_metaclass(abc.ABCMeta)
class AbstractExpression(object):

    def __init__(self, *clauses):
        super(AbstractExpression, self).__init__()
        self._clauses = clauses

    def extend_clauses(self, clauses):
        self._clauses = self._clauses + clauses

    @property
    def clauses(self):
        return self._clauses

    @property
    def value(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def construct_expression(self):
        raise NotImplementedError()


class ClauseList(AbstractExpression):
    @property
    def value(self):
        res = []
        for val in self._clauses:
            if isinstance(val, AbstractExpression):
                res.extend(val.value)
            else:
                res.append(val.value)
        return res

    @property
    def operator(self):
        raise NotImplementedError

    def construct_expression(self):
        return (
            "("
            + (" " + self.operator + " ").join(
                val.construct_expression() for val in self._clauses)
            + ")"
        ) if self._clauses else ""


class AND(ClauseList):
    operator = "AND"


class OR(ClauseList):
    operator = "OR"


FILTER_MAPPING = {
    filters.EQ: EQ,
    filters.NE: NE,
    filters.GT: GT,
    filters.GE: GE,
    filters.LE: LE,
    filters.LT: LT,
    filters.Is: Is,
    filters.IsNot: IsNot,
    filters.In: In,
    filters.NotIn: NotIn,
    filters.Like: Like,
    filters.NotLike: NotLike,
}

FILTER_EXPR_MAPPING = {
    filters.AND: AND,
    filters.OR: OR,
}


class AsIsType(types.BaseType):

    def validate(self, value):
        return True

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return value


def convert_filters(model, filters_root):
    filters_root = filters_root or filters.AND()
    if isinstance(filters_root, filters.AbstractExpression):
        return iterate_filters(model, filters_root)
    return iterate_filters(model, filters.AND(filters_root))


def iterate_filters(model, filter_list):
    # Just expression
    if isinstance(filter_list, filters.AbstractExpression):
        clauses = iterate_filters(model, filter_list.clauses)
        return FILTER_EXPR_MAPPING[type(filter_list)](*clauses)

    # Tuple of causes from expression
    if isinstance(filter_list, tuple):
        clauses = []
        for cause in filter_list:
            c_causes = iterate_filters(model, cause)
            if isinstance(cause, filters.AbstractExpression):
                clauses.append(c_causes)
            else:
                clauses.extend(c_causes)
        return clauses

    # old style mappings (dict, multidict)
    if isinstance(filter_list, collections_abc.Mapping):
        clauses = []
        for name, filt in filter_list.items():
            if isinstance(model, common.TableAlias):
                value_type = (model.original.model.properties.properties[name]
                              .get_property_type()) or AsIsType()
                column = model.get_column_by_name(
                    name, wrap_alias=False,
                )
            else:
                value_type = (model.properties.properties[name]
                              .get_property_type()) or AsIsType()
                column = utils.escape(name)
            # Make API compatible with previous versions.
            if not isinstance(filt, filters.AbstractClause):
                LOG.warning("DEPRECATED: pleases use %s wrapper for filter "
                            "value", filters.EQ)
                clauses.append(EQ(column, value_type, filt))
                continue

            try:
                clauses.append(FILTER_MAPPING[type(filt)](
                    column, value_type, filt.value))
            except KeyError:
                raise ValueError("Can't convert API filter to SQL storage "
                                 "filter. Unknown filter %s" % filt)
        return clauses

    raise ValueError("Unknown type of filters: %s" % filter_list)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2018 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import logging
from six.moves import http_client

from restalchemy.api import middlewares
from restalchemy.common import exceptions as comm_exc
from restalchemy.storage import exceptions as ra_exceptions

UNKNOWN_ERROR_CODE = 10000001


LOG = logging.getLogger(__name__)


def exception2dict(exception):
    code = (exception.get_code() if hasattr(exception, 'get_code') else
            exception.code if hasattr(exception, 'code') else
            UNKNOWN_ERROR_CODE)
    message = (exception.msg if hasattr(exception, 'msg') else
               exception.message if hasattr(exception, 'message') else
               str(exception))
    return {
        'type': exception.__class__.__name__,
        'code': code,
        'message': message
    }


class ErrorsHandlerMiddleware(middlewares.Middleware):

    not_found_exc = (ra_exceptions.RecordNotFound,)
    conflict_exc = (ra_exceptions.ConflictRecords,)
    common_exc = (comm_exc.RestAlchemyException,)
    valid_exc = (comm_exc.ValidationErrorException,)

    def process_request(self, req):
        try:
            return req.get_response(self.application)
        except self.not_found_exc as e:
            return req.ResponseClass(status=http_client.NOT_FOUND,
                                     json=exception2dict(e))
        except self.conflict_exc as e:
            return req.ResponseClass(status=http_client.CONFLICT,
                                     json=exception2dict(e))
        except self.valid_exc as e:
            return req.ResponseClass(status=http_client.BAD_REQUEST,
                                     json=exception2dict(e))
        except self.common_exc as e:
            return req.ResponseClass(status=e.code,
                                     json=exception2dict(e))
        except Exception as e:
            LOG.exception("Unknown error has occurred on url: %s %s",
                          req.method, req.url)
            return req.ResponseClass(status=http_client.INTERNAL_SERVER_ERROR,
                                     json=exception2dict(e))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#    Copyright 2023 v.burygin
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
from wsgiref.simple_server import make_server

from restalchemy.api import applications
from restalchemy.api import controllers
from restalchemy.api import middlewares
from restalchemy.api import routes

from restalchemy.openapi import engines as openapi_engines
from restalchemy.openapi import structures as openapi_structures

HOST = '0.0.0.0'
PORT = 8000


class RootController(controllers.Controller):
    """Controller for / endpoint"""

    def filter(self, filters):
        return ["v1"]


class ApiEndpointRoute(routes.Route):
    """Handler for /v1/ endpoint"""

    __controller__ = controllers.RootController
    __allow_methods__ = [
        routes.FILTER,
        routes.GET,
    ]

    specifications = routes.action(routes.OpenApiSpecificationRoute)


class UserApiApp(routes.Route):
    __controller__ = controllers.RootController
    __allow_methods__ = [routes.FILTER]


# Route to /v1/ endpoint.
setattr(
    UserApiApp,
    "v1",
    routes.route(ApiEndpointRoute),
)


def get_openapi_engine():
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
    )
    return openapi_engine


def get_user_api_application():
    return UserApiApp


def build_wsgi_application():
    return middlewares.attach_middlewares(
        applications.OpenApiApplication(route_class=get_user_api_application(),
                                        openapi_engine=get_openapi_engine()),
        [],
    )


def main():
    """

    After start you can try curl http://127.0.0.1:8000/v1/specifications/3.0.3

    """
    server = make_server(HOST, PORT, build_wsgi_application())

    try:
        six.print_("Serve forever on %s:%s" % (HOST, PORT))
        server.serve_forever()
    except KeyboardInterrupt:
        six.print_('Bye')


if __name__ == '__main__':
    main()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import  # noqa

import abc

from mysql.connector import errors
import six

from restalchemy.storage.sql.dialect import base
from restalchemy.storage.sql.dialect import exceptions as exc
from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.storage.sql import filters as sql_filters
from restalchemy.storage.sql import utils


@six.add_metaclass(abc.ABCMeta)
class AbstractProcessResult(base.AbstractProcessResult):

    def get_count(self):
        return self._result.rowcount

    @property
    def rows(self):
        return self.get_rows()

    @abc.abstractmethod
    def fetchall(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_rows(self):
        raise NotImplementedError()


class MySQLProcessResult(AbstractProcessResult):

    def __init__(self, result):
        super(MySQLProcessResult, self).__init__(result)
        self._rows = None

    def fetchall(self):
        for row in self._result:
            yield row

    def get_rows(self):
        if self._rows is None:
            self._rows = self._result.fetchall()
        return self._rows


class MySqlOrmProcessResult(AbstractProcessResult):

    def __init__(self, result, query):
        super(MySqlOrmProcessResult, self).__init__(result=result)
        self._query = query
        self._rows = None

    def fetchall(self):
        for row in self._result:
            yield self._query.parse_row(row)

    def get_rows(self):
        if self._rows is None:
            self._rows = self._query.parse_results(self._result.fetchall())
        return self._rows


@six.add_metaclass(abc.ABCMeta)
class AbstractDialectCommand(base.AbstractDialectCommand):

    def execute(self, session):
        try:
            return MySQLProcessResult(
                super(AbstractDialectCommand, self).execute(session))
        except errors.DatabaseError as e:
            if e.errno == 1213:
                raise exc.DeadLock(code=e.sqlstate, message=e.msg)
            elif e.errno == 1062:
                raise exc.Conflict(code=e.sqlstate, message=e.msg)
            raise


class MySQLInsert(AbstractDialectCommand):

    def get_values(self):
        values = tuple()
        for column_name in self._table.get_column_names():
            values += (self._data[column_name],)
        return values

    def get_statement(self):
        column_names = self._table.get_escaped_column_names()
        return "INSERT INTO `%s` (%s) VALUES (%s)" % (
            self._table.name,
            ", ".join(column_names),
            ", ".join(['%s'] * len(column_names))
        )


class MySQLUpdate(AbstractDialectCommand):

    def __init__(self, table, ids, data):
        super(MySQLUpdate, self).__init__(table, data)
        self._ids = ids

    def get_values(self):
        values = tuple()
        column_names = self._table.get_column_names(with_pk=False)
        pk_names = self._table.get_pk_names()
        for column_name in column_names:
            values += (self._data[column_name],)
        for column_name in pk_names:
            values += (self._ids[column_name],)
        return values

    def get_statement(self):
        column_names = self._table.get_escaped_column_names(with_pk=False)
        pk_names = self._table.get_escaped_pk_names()
        return "UPDATE `%s` SET %s WHERE %s" % (
            self._table.name,
            ", ".join(["%s = %s" % (name, "%s") for name in column_names]),
            " AND ".join(["%s = %s" % (name, "%s") for name in pk_names])
        )


class MySQLDelete(AbstractDialectCommand):

    def __init__(self, table, ids):
        super(MySQLDelete, self).__init__(table=table, data={})
        self._ids = ids

    def get_values(self):
        values = tuple()
        pk_names = self._table.get_pk_names()
        for column_name in pk_names:
            values += (self._ids[column_name],)
        return values

    def get_statement(self):
        pk_names = self._table.get_escaped_pk_names()
        return "DELETE FROM `%s` WHERE %s" % (
            self._table.name,
            " AND ".join(["%s = %s" % (name, "%s") for name in pk_names])
        )


class MySQLBatchDelete(AbstractDialectCommand):

    def __init__(self, table, snapshot):
        super(MySQLBatchDelete, self).__init__(table=table, data={})
        self._snapshot = snapshot
        self._pk_keys = self._table.get_escaped_pk_names()
        keys_count = len(self._pk_keys)
        if keys_count == 1:
            self._is_multiple_primary_key = False
        elif keys_count > 1:
            self._is_multiple_primary_key = True
        else:
            raise ValueError("The model with table %r has 0 primary keys" %
                             table)

    def _get_values(self):
        values = []
        for snapshot in self._snapshot:
            for key in self._table.get_pk_names():
                values.append(snapshot[key])
        return values

    def _get_multiple_primary_key_values(self):
        return self._get_values()

    def _get_single_primary_key_values(self):
        # NOTE(efrolov): Wrap to list for `in` optimization
        return [self._get_multiple_primary_key_values()]

    def get_values(self):
        return (self._get_multiple_primary_key_values()
                if self._is_multiple_primary_key else
                self._get_single_primary_key_values())

    def _get_single_primary_key_statement(self):
        return "DELETE FROM `%s` WHERE %s in %s" % (
            self._table.name,
            self._pk_keys[0],
            "%s"
        )

    def _get_multiple_primary_key_statement(self):
        where_part = " AND ".join(
            [("%s = %s" % (key, '%s')) for key in self._pk_keys]
        )
        where_condition = " OR ".join(
            [where_part for _ in range(len(self._snapshot))]
        )
        return "DELETE FROM `%s` WHERE %s" % (
            self._table.name,
            where_condition,
        )

    def get_statement(self):
        return (self._get_multiple_primary_key_statement()
                if self._is_multiple_primary_key else
                self._get_single_primary_key_statement())


class MySQLBasicSelect(AbstractDialectCommand):
    def __init__(self, table, limit=None, order_by=None, locked=False):
        super(MySQLBasicSelect, self).__init__(table=table, data={})
        self._limit = limit
        self._order_by = order_by
        self._locked = locked

    def construct_limit(self):
        if self._limit:
            return " LIMIT " + str(self._limit)
        return ""

    def construct_locked(self):
        if self._locked:
            return " FOR UPDATE"
        return ""

    def construct_order_by(self):
        if self._order_by:
            res = []
            for name, sorttype in self._order_by.items():
                sorttype = sorttype.upper()
                if sorttype not in ['ASC', 'DESC', '', None]:
                    raise ValueError("Unknown order: %s." % sorttype)
                res.append('%s %s' % (utils.escape(name), sorttype or 'ASC'))
            return " ORDER BY " + ", ".join(res)
        return ""


class MySQLSelect(MySQLBasicSelect):

    def __init__(self, table, filters=None, limit=None,
                 order_by=None, locked=False):
        super(MySQLSelect, self).__init__(
            table=table, limit=limit, order_by=order_by, locked=locked)
        self._filters = sql_filters.convert_filters(self._table.model, filters)

    def get_values(self):
        return self._filters.value

    def construct_where(self):
        return self._filters.construct_expression()

    def get_statement(self):
        sql = "SELECT %s FROM `%s`" % (
            ", ".join(self._table.get_escaped_column_names()),
            self._table.name
        )
        filt = self.construct_where()

        return (sql + (" WHERE %s" % filt if filt else "")
                + self.construct_order_by()
                + self.construct_limit()
                + self.construct_locked())


class MySQLCustomSelect(MySQLBasicSelect):

    def __init__(self, table, where_conditions, where_values, limit=None,
                 order_by=None, locked=False):
        super(MySQLCustomSelect, self).__init__(
            table=table, limit=limit, order_by=order_by, locked=locked)
        self._where_conditions = where_conditions
        self._where_values = where_values

    def get_values(self):
        return self._where_values

    def construct_where(self):
        return self._where_conditions

    def get_statement(self):
        sql = "SELECT %s FROM `%s`" % (
            ", ".join(self._table.get_escaped_column_names()),
            self._table.name
        )
        return sql + " WHERE " + self.construct_where() \
            + self.construct_order_by() + self.construct_limit() \
            + self.construct_locked()


class MySQLCount(MySQLSelect):

    def __init__(self, table, filters=None):
        super(MySQLCount, self).__init__(table=table, filters=filters)

    def get_statement(self):
        sql = "SELECT COUNT(*) as COUNT FROM `%s`" % (
            self._table.name
        )
        filt = self.construct_where()

        return sql + (" WHERE %s" % filt if filt else "")


class MySqlOrmDialectCommand(base.AbstractDialectCommand):

    def __init__(self, table, query):
        super(MySqlOrmDialectCommand, self).__init__(table=table,
                                                     data=None)
        self._query = query

    def get_statement(self):
        return self._query.compile()

    def get_values(self):
        return self._query.values()

    def execute(self, session):
        try:
            return MySqlOrmProcessResult(
                result=super(MySqlOrmDialectCommand, self).execute(session),
                query=self._query,
            )
        except errors.DatabaseError as e:
            if e.errno == 1213:
                raise exc.DeadLock(code=e.sqlstate, message=e.msg)
            elif e.errno == 1062:
                raise exc.Conflict(code=e.sqlstate, message=e.msg)
            raise


class MySqlOrm(object):

    @staticmethod
    def select(model):
        return q.Q.select(model)


class MySQLDialect(base.AbstractDialect):

    def __init__(self):
        super(MySQLDialect, self).__init__()
        self._orm = MySqlOrm()

    @property
    def orm(self):
        return self._orm

    def orm_command(self, table, query):
        return MySqlOrmDialectCommand(table, query)

    def insert(self, table, data):
        return MySQLInsert(table, data)

    def update(self, table, ids, data):
        return MySQLUpdate(table, ids, data)

    def delete(self, table, ids):
        return MySQLDelete(table, ids)

    def select(self, table, filters, limit=None, order_by=None, locked=False):
        return MySQLSelect(table, filters, limit, order_by, locked)

    def custom_select(self, table, where_conditions, where_values, limit=None,
                      order_by=None, locked=False):
        return MySQLCustomSelect(table, where_conditions, where_values, limit,
                                 order_by, locked)

    def count(self, table, filters):
        return MySQLCount(table, filters)



# coding=utf-8
#
# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
import unittest

import six

from restalchemy.dm import types
from restalchemy.tests.functional import base


@unittest.skipUnless(six.PY2, "Actual only for python2")
class ThreadImportTestCase(base.BaseFunctionalTestCase):

    def test_strptime(self):
        exc_raised = [0]

        def f(exc_raised):
            try:
                for _ in range(1, 13):
                    for _ in range(1, 29):
                        types.UTCDateTime().from_simple_type(
                            '2013-01-01 00:00:00.0')
            except AttributeError:
                exc_raised[0] += 1

        pool = []
        for _ in range(10):
            thread = threading.Thread(target=f, args=(exc_raised,))
            pool.append(thread)
            thread.start()

        for thread in pool:
            thread.join()

        self.assertEqual(0, exc_raised[0])



# coding=utf-8
# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base


class FakeModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "batch_insert"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


class TestOrderByTestCase(base.BaseWithDbMigrationsTestCase):

    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_without_order_by(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2])

        all_models = set(FakeModel.objects.get_all())

        self.assertEqual({model1, model2}, all_models)

    def test_with_order_by_asc(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2])

        all_models = FakeModel.objects.get_all(
            order_by={'foo_field1': 'ASC'})

        self.assertEqual([model1, model2], all_models)

    def test_with_order_by_desc(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2])

        all_models = FakeModel.objects.get_all(
            order_by={'foo_field1': 'DESC'})

        self.assertEqual([model2, model1], all_models)


class TestLikeTestCase(base.BaseWithDbMigrationsTestCase):

    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_filter_like_not_like(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")
        model3 = FakeModel(foo_field1=3, foo_field2="FooModel3")
        model4 = FakeModel(foo_field1=4, foo_field2="TestMod-4")
        model5 = FakeModel(foo_field1=5, foo_field2="Mod'el")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2, model3, model4, model5])

        # Like: Startswith Model
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like("Model%")})
        )

        self.assertEqual({model1, model2}, all_models)

        # Like: del in center
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like("%del%")})
        )

        self.assertEqual({model1, model2, model3}, all_models)

        # Like: Endswith 3
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like("%3")})
        )

        self.assertEqual({model3, }, all_models)

        # Like: TestMod + random symbol + 4
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like("TestMod_4")})
        )

        self.assertEqual({model4, }, all_models)

        # Like: Endswith 5
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like("%5")})
        )

        self.assertEqual(set(), all_models)

        # Like: Startswith Foo or Test
        # TODO(v.burygin) Зачем тут AND, - неизвестно, надо разбираться
        filter_list = dm_filters.AND(dm_filters.OR(
            {'foo_field2': dm_filters.Like("Foo%")},
            {'foo_field2': dm_filters.Like("Test%")}))
        all_models = set(FakeModel.objects.get_all(
            filters=filter_list)
        )

        self.assertEqual({model3, model4}, all_models)

        # Like: ' in center with escape symbol
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.Like(r"Mod\'el")})
        )

        self.assertEqual({model5, }, all_models)

        # NotLike: Not startswith Foo
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.NotLike("Foo%")})
        )

        self.assertEqual({model1, model2, model4, model5}, all_models)

        # NotLike: Mod in center
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.NotLike("%Mod%")})
        )

        self.assertEqual(set(), all_models)

        # NotLike: ' in center with escape symbol
        all_models = set(FakeModel.objects.get_all(
            filters={"foo_field2": dm_filters.NotLike(r"Mod\'el")})
        )

        self.assertEqual({model1, model2, model3, model4}, all_models)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstarctMigrationStep):

    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "e31a12bb-3c3a-4f86-8bdd-ca9f7b613b6c"

    def upgrade(self, session):
        expressions = [
            """
                CREATE TABLE IF NOT EXISTS vms (
                    uuid CHAR(36) NOT NULL,
                    state VARCHAR(10) NOT NULL,
                    name VARCHAR(255) NOT NULL,
                    just_none VARCHAR(255) NULL,
                    status VARCHAR(255) NULL,
                    `created` DATETIME(6) DEFAULT NOW(),
                    `updated` DATETIME(6) DEFAULT NOW(),
                    PRIMARY KEY (uuid)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
            """, """
                CREATE TABLE IF NOT EXISTS ports (
                    uuid CHAR(36) NOT NULL,
                    mac CHAR(17) NOT NULL,
                    vm CHAR(36) NOT NULL,
                    PRIMARY KEY (uuid),
                    CONSTRAINT FOREIGN KEY ix_vms_uuid (vm)
                    REFERENCES vms (uuid)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
            """, """
                CREATE TABLE IF NOT EXISTS ip_addresses (
                    uuid CHAR(36) NOT NULL,
                    ip CHAR(17) NOT NULL,
                    port CHAR(36) NOT NULL,
                    PRIMARY KEY (uuid),
                    CONSTRAINT FOREIGN KEY ix_ports_uuid (port)
                    REFERENCES ports (uuid)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
            """, """
                CREATE TABLE IF NOT EXISTS tags (
                    uuid CHAR(36) NOT NULL,
                    vm CHAR(36) NOT NULL,
                    name CHAR(40) NOT NULL,
                    visible BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    CONSTRAINT FOREIGN KEY ix_tags__vms_uuid (vm)
                    REFERENCES vms (uuid)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
            """
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        tables = ['tags', 'ip_addresses', 'ports', 'vms']

        for table in tables:
            self._delete_table_if_exists(session, table)


migration_step = MigrationStep()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class AbstractProcessResult(object):

    def __init__(self, result):
        self._result = result

    @abc.abstractmethod
    def get_count(self):
        raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class AbstractDialectCommand(object):

    def __init__(self, table, data):
        self._table = table
        self._data = data

    @abc.abstractmethod
    def get_values(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_statement(self):
        raise NotImplementedError()

    def execute(self, session):
        values = self.get_values()
        statement = self.get_statement()
        return session.execute(statement, values)


@six.add_metaclass(abc.ABCMeta)
class AbstractDialect(object):

    @abc.abstractproperty
    def insert(self, table, data):
        raise NotImplementedError()

    @abc.abstractproperty
    def update(self, table, ids, data):
        raise NotImplementedError()

    @abc.abstractproperty
    def delete(self, table, ids):
        raise NotImplementedError()

    @abc.abstractproperty
    def select(self, table, filters, limit=None, order_by=None, locked=False):
        raise NotImplementedError()

    @abc.abstractproperty
    def count(self, table, filters):
        raise NotImplementedError()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2021 George Melikov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from restalchemy.storage.sql.dialect.query_builder import common
from restalchemy.storage.sql.dialect.query_builder import q


class TestOrderByValue(unittest.TestCase):
    def setUp(self):
        self.column = common.Column('1', None)

    def test_empty_type(self):
        order = q.OrderByValue(self.column)

        self.assertEqual('`1` ASC', order.compile())

    def test_valid_type(self):
        order = q.OrderByValue(self.column, sort_type='DESC')

        self.assertEqual('`1` DESC', order.compile())

    def test_invalid_type(self):
        with self.assertRaises(ValueError):
            q.OrderByValue(self.column, sort_type='WRONG')

    def test_valid_type_lowercase(self):
        order = q.OrderByValue(self.column, sort_type='desc')

        self.assertEqual('`1` DESC', order.compile())



# coding=utf-8
#
# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import re
import uuid

import mock
import six

from restalchemy.dm import types
from restalchemy.tests.unit import base


TEST_STR_VALUE = 'test_value :)'
TEST_INT_AS_STR_VALUE = '1234'
TEST_INT_VALUE = 5
TEST_TYPE = 'FAKE TYPE'
INCORECT_UUID = '4a775g98-eg85-4a0e-a0g0-639f0a16f4c3'
INCORECT_INT_AS_STR = '123abc'


@mock.patch("re.compile", return_value=mock.MagicMock(), autospec=True)
class BaseRegExpTypeTestCase(base.BaseTestCase):

    def _prepare_mock(self, re_mock, return_value):
        self.re_match_mock = mock.MagicMock(**{
            'match': mock.MagicMock(return_value=return_value)})
        re_mock.return_value = self.re_match_mock

    def test_correct_value_if_value_is_not_none(self, re_mock):
        self._prepare_mock(re_mock, re.match("a", "a"))

        test_instance = types.BaseRegExpType("")

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))
        self.re_match_mock.match.assert_called_once_with(TEST_STR_VALUE)

    def test_correct_value_if_value_is_none(self, re_mock):
        self._prepare_mock(re_mock, None)

        test_instance = types.BaseRegExpType("")

        self.assertFalse(test_instance.validate(None))

    def test_incorrect_value(self, re_mock):
        self._prepare_mock(re_mock, None)

        test_instance = types.BaseRegExpType("")

        self.assertFalse(test_instance.validate(TEST_STR_VALUE))
        self.re_match_mock.match.assert_called_once_with(TEST_STR_VALUE)


class BaseCompiledRegExpTypeTestCase(base.BaseTestCase):

    def test_correct_value_if_value_is_not_none(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r"t"))

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))

    def test_correct_value_if_value_is_none(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r""))

        self.assertFalse(test_instance.validate(None))

    def test_incorrect_value(self):
        test_instance = types.BaseCompiledRegExpType(re.compile(r"-"))

        self.assertFalse(test_instance.validate(TEST_STR_VALUE))

    def test_get_pattern_from_attribute(self):
        class TestType(types.BaseCompiledRegExpTypeFromAttr):
            pattern = re.compile(r"t")
        test_instance = TestType()

        self.assertTrue(test_instance.validate(TEST_STR_VALUE))


class BaseTestCase(base.BaseTestCase):

    def __init__(self, *args, **kwargs):
        super(BaseTestCase, self).__init__(*args, **kwargs)
        self.test_instance = mock.MagicMock()
        self.test_instance.validate.configure_mock(**{'return_value': False})

    def test_correct_none_value(self):
        self.assertFalse(self.test_instance.validate(None))


class UUIDTestCase(base.BaseTestCase):

    def setUp(self):
        super(UUIDTestCase, self).setUp()
        self.test_instance = types.UUID()

    def test_uuid_correct_value(self):
        self.assertTrue(self.test_instance.validate(uuid.uuid4()))

    def test_uuid_incorrect_value(self):
        INCORECT_UUID = '4a775g98-eg85-4a0e-a0g0-639f0a16f4c3'

        self.assertFalse(self.test_instance.validate(
            INCORECT_UUID))

    def test_to_simple_type(self):
        TEST_UUID = uuid.uuid4()

        self.assertEqual(
            self.test_instance.to_simple_type(TEST_UUID),
            str(TEST_UUID))

    def test_from_simple_type(self):
        TEST_UUID = uuid.uuid4()

        self.assertEqual(
            self.test_instance.from_simple_type(str(TEST_UUID)),
            TEST_UUID)


class StringTestCase(base.BaseTestCase):

    FAKE_STRING1 = 'fake!!!'
    FAKE_STRING2 = six.u('fake!!!')

    def setUp(self):
        super(StringTestCase, self).setUp()
        self.test_instance1 = types.String(min_length=5, max_length=8)
        self.test_instance2 = types.String()

    def test_correct_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING1))

    def test_correct_unicode_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING2))

    def test_correct_min_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING1[:5]))

    def test_correct_min_unicode_value(self):
        self.assertTrue(self.test_instance1.validate(self.FAKE_STRING2[:5]))

    def test_correct_max_value(self):
        self.assertTrue(self.test_instance1.validate(
            (self.FAKE_STRING1 * 2)[:8]))

    def test_correct_max_unicode_value(self):
        self.assertTrue(self.test_instance1.validate(
            (self.FAKE_STRING2 * 2)[:8]))

    def test_incorrect_min_value(self):
        self.assertFalse(self.test_instance1.validate(self.FAKE_STRING1[:4]))

    def test_incorrect_min_unicode_value(self):
        self.assertFalse(self.test_instance1.validate(self.FAKE_STRING1[:4]))

    def test_incorrect_max_value(self):
        self.assertFalse(self.test_instance1.validate(
            (self.FAKE_STRING1 * 2)[:9]))

    def test_incorrect_max_unicode_value(self):
        self.assertFalse(self.test_instance1.validate(
            (self.FAKE_STRING1 * 2)[:9]))

    def test_correct_infinity_value(self):
        self.assertTrue(self.test_instance2.validate(
            self.FAKE_STRING1 * 100500))

    def test_incorrect_type_validate(self):
        self.assertFalse(self.test_instance1.validate(5))


class IntegerTestCase(base.BaseTestCase):

    def setUp(self):
        super(IntegerTestCase, self).setUp()

        self.test_instance = types.Integer(0, 55)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(30))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(55))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(0))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST_STR_VALUE"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(56))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(-1))

    def test_validate_sys_max_value(self):
        test_instance = types.Integer()

        self.assertTrue(test_instance.validate(six.MAXSIZE))

    def test_validate_sys_min_value(self):
        test_instance = types.Integer()

        self.assertTrue(test_instance.validate(-six.MAXSIZE))


class FloatTestCase(base.BaseTestCase):

    def setUp(self):
        super(FloatTestCase, self).setUp()

        self.test_instance = types.Float(0.0, 55.0)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(30.0))

    def test_validate_correct_max_value(self):
        self.assertTrue(self.test_instance.validate(55.0))

    def test_validate_correct_min_value(self):
        self.assertTrue(self.test_instance.validate(0.0))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate("TEST_STR_VALUE"))

    def test_validate_incorrect_max_value(self):
        self.assertFalse(self.test_instance.validate(56.0))

    def test_validate_incorrect_min_value(self):
        self.assertFalse(self.test_instance.validate(-1.0))

    def test_validate_sys_max_value(self):
        test_instance = types.Float()

        self.assertTrue(test_instance.validate(float(six.MAXSIZE)))

    def test_validate_sys_min_value(self):
        test_instance = types.Float()

        self.assertTrue(test_instance.validate(float(-six.MAXSIZE)))


class UriTestCase(BaseTestCase):

    def setUp(self):
        super(UriTestCase, self).setUp()
        self.test_instance = types.Uri()

    def test_correct_value(self):
        self.assertTrue(self.test_instance.validate(
            '/fake/fake/' + str(uuid.uuid4())))

    def test_incorect_uuid_value(self):
        self.assertFalse(self.test_instance.validate(
            '/fake/fake/' + INCORECT_UUID))

    def test_incorect_start_char_value(self):
        self.assertFalse(self.test_instance.validate(
            'fake/fake/' + str(uuid.uuid4())))

    def test_incorect_start_end_value(self):
        self.assertFalse(self.test_instance.validate(
            '/fake/fake' + str(uuid.uuid4())))


class MacTestCase(BaseTestCase):

    def setUp(self):
        super(MacTestCase, self).setUp()
        self.test_instance = types.Mac()

    def get_values(self, value):
        return [value, value.upper()]

    def test_correct_value(self):
        for value in self.get_values("05:06:07:08:ab:ff"):
            self.assertTrue(self.test_instance.validate(value))

    def test_incorrect_cahar_value(self):
        for value in self.get_values("05:06:0k:08:ab:ff"):
            self.assertFalse(self.test_instance.validate(value))

    def test_incorrect_length_value(self):
        for value in self.get_values("05:06:08:ab:ff"):
            self.assertFalse(self.test_instance.validate(value))


class BasePythonTypeTestCase(base.BaseTestCase):

    def setUp(self):
        super(BasePythonTypeTestCase, self).setUp()

        self.test_instance = types.BasePythonType(int)

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(TEST_INT_VALUE))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class ListTestCase(base.BaseTestCase):

    def setUp(self):
        super(ListTestCase, self).setUp()

        self.test_instance = types.List()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(list()))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class TypedListTestCase(base.BaseTestCase):

    def setUp(self):
        super(TypedListTestCase, self).setUp()

        self.test_instance_int = types.TypedList(nested_type=types.Integer())
        self.test_instance_str = types.TypedList(nested_type=types.String())

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance_int.validate([]))
        self.assertTrue(self.test_instance_int.validate([1, 2, 3]))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance_int.validate([1, 2, '3', 4]))

    def test_incorrect_nested_type(self):
        self.assertRaises(TypeError, types.TypedList, int)

    def test_from_unicode(self):
        payload = (
            (TEST_INT_AS_STR_VALUE, [int(TEST_INT_AS_STR_VALUE)]),
        )

        for value, expectedResult in payload:
            result = self.test_instance_int.from_unicode(value)
            self.assertEqual(expectedResult, result)

        payload = (
            (TEST_STR_VALUE, [TEST_STR_VALUE]),
        )

        for value, expectedResult in payload:
            result = self.test_instance_str.from_unicode(value)
            self.assertEqual(expectedResult, result)

        with self.assertRaises(TypeError):
            self.test_instance_int.from_unicode(None)
            self.test_instance_int.from_unicode(INCORECT_INT_AS_STR)


class DictTestCase(base.BaseTestCase):

    def setUp(self):
        super(DictTestCase, self).setUp()

        self.test_instance = types.Dict()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(dict()))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))


class TypedDictTestCase(base.BaseTestCase):

    def setUp(self):
        super(TypedDictTestCase, self).setUp()

        self.scheme_simple_types = {
            'int': types.Integer(),
            'str': types.String(),
        }
        self.scheme_lists = {
            'list': types.List(),
            'typed_list': types.TypedList(types.Integer()),
        }
        self.scheme_dicts = {
            'dict': types.Dict(),
            'typed_dict': types.SchemeDict({
                'sub_str': types.String(),
                'sub_int': types.Integer(),
            }),
        }
        self.scheme_dict_sublist = {
            'typed_dict_with_typed_list': types.SchemeDict(
                {'sub_list_typed': types.TypedList(types.String())}),
        }
        self.scheme_dict_subdict = {
            'typed_dict_with_typed_dict': types.SchemeDict(
                {'sub_dict_typed': types.SchemeDict(
                    {
                        'sub_str': types.String(),
                        'sub_int': types.Integer(),
                    })}),
        }

    def test_schema_keys_not_string(self):
        self.assertRaises(ValueError, types.SchemeDict, {1: types.Integer()})

    def test_schema_values_not_types(self):
        self.assertRaises(ValueError, types.SchemeDict, {'1': int})

    def test_validate_simple_schema(self):
        dict_type = types.SchemeDict(scheme=self.scheme_simple_types)

        self.assertTrue(dict_type.validate({'int': 1, 'str': 'string'}))

    def test_validate_simple_schema_missing_item(self):
        dict_type = types.SchemeDict(scheme=self.scheme_simple_types)

        self.assertFalse(dict_type.validate({'int': 1}))
        self.assertFalse(dict_type.validate({'str': 'string'}))

    def test_validate_simple_schema_extra_item(self):
        dict_type = types.SchemeDict(scheme={'int': types.Integer()})

        self.assertFalse(dict_type.validate({'int': 1, 'str': 'string'}))

    def test_validate_simple_schema_invalid_value(self):
        dict_type_1 = types.SchemeDict(scheme={'int': types.Integer()})
        dict_type_2 = types.SchemeDict(scheme={'str': types.String()})

        self.assertFalse(dict_type_1.validate({'int': '1'}))
        self.assertFalse(dict_type_2.validate({'str': None}))

    def test_validate_schema_with_lists(self):
        schema = {'mixed_list': types.List()}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_lists)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate({'int': 1,
                                'str': 'string',
                                'list': [1, 2, 3],
                                'mixed_list': [1, 'a', None],
                                'typed_list': [1, 2, 3]}))

    def test_validate_schema_incorrect_typed_list_value(self):
        schema = {'typed_list': types.TypedList(types.Integer())}

        dict_type = types.SchemeDict(scheme=schema)

        self.assertFalse(dict_type.validate({'typed_list': [1, '2', 3]}))
        self.assertFalse(dict_type.validate({'typed_list': [None, 2, 3]}))
        self.assertFalse(dict_type.validate({'typed_list': [1, 2, {}]}))

    def test_validate_schema_with_dicts(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_dicts)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate({'int': 1,
                                'str': 'string',
                                'dict': {'1': 1, '2': 'a', 'z': 3},
                                'typed_dict': {
                                    'sub_str': 'string',
                                    'sub_int': 42,
                                }}))

    def test_validate_schema_subdict_missing_item(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_dicts)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertFalse(
            dict_type.validate({'int': 1,
                                'str': 'string',
                                'dict': {1: 1, 2: 'a', 'z': 3},
                                'typed_dict': {
                                    'sub_str': 'string',
                                }}))

    def test_validate_complex_schema(self):
        schema = {}
        schema.update(self.scheme_simple_types)
        schema.update(self.scheme_lists)
        schema.update(self.scheme_dicts)
        schema.update(self.scheme_dict_sublist)
        schema.update(self.scheme_dict_subdict)

        dict_type = types.SchemeDict(scheme=schema)

        self.assertTrue(
            dict_type.validate({'int': 1,
                                'str': 'string',
                                'list': ['a'],
                                'typed_list': [],
                                'dict': {},
                                'typed_dict': {
                                    'sub_str': 'string',
                                    'sub_int': 42},
                                'typed_dict_with_typed_list': {
                                    'sub_list_typed': ['s']
                                },
                                'typed_dict_with_typed_dict': {
                                    'sub_dict_typed': {
                                        'sub_str': 'string_2',
                                        'sub_int': -5}
                                },
                                }))

    def test_scheme_dict_to_simple_type(self):
        val = 7
        ret_val = 9
        test_obj = self

        class TestType(types.Integer):
            def to_simple_type(self, data_val):
                test_obj.assertIs(val, data_val)
                return ret_val

        scheme = {
            'key': TestType(),
        }
        data = {
            'key': val,
        }
        expect_data = {
            'key': ret_val
        }

        res = types.SchemeDict(scheme=scheme).to_simple_type(data)
        self.assertEqual(expect_data, res)


class UTCDateTimeTestCase(base.BaseTestCase):

    def setUp(self):
        super(UTCDateTimeTestCase, self).setUp()

        self.test_instance = types.UTCDateTime()

    def test_validate_correct_value(self):
        self.assertTrue(
            self.test_instance.validate(datetime.datetime.utcnow()))

    def test_validate_incorrect_value_type(self):
        self.assertFalse(self.test_instance.validate(TEST_STR_VALUE))

    def test_validate_incorrect_value_tzinfo(self):
        self.assertFalse(
            self.test_instance.validate(
                datetime.datetime.utcnow().replace(tzinfo=datetime.tzinfo())))

    def test_zero_microseconds(self):
        dt = datetime.datetime(2020, 3, 13, 11, 3, 25)
        expected = '2020-03-13 11:03:25.000000'
        dt_type = types.UTCDateTime()

        result = dt_type.to_simple_type(dt)

        self.assertEqual(result, expected)

    def test_openapi_format(self):
        dt = datetime.datetime(2020, 3, 13, 11, 3, 25, 123)

        expected = '2020-03-13T11:03:25.000123Z'
        dt_type = types.UTCDateTime()

        result = dt_type.dump_value(dt)

        self.assertEqual(result, expected)

        result = dt_type.from_simple_type(expected)
        self.assertEqual(result, dt)


class EnumTestCase(base.BaseTestCase):

    def setUp(self):
        super(EnumTestCase, self).setUp()

        self.test_instance = types.Enum([1, 2, 3])

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(1))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(4))


class AllowNoneTestCase(base.BaseTestCase):

    def setUp(self):
        super(AllowNoneTestCase, self).setUp()

        self.test_instance = types.AllowNone(types.String())

    def test_from_unicode(self):
        s = "blah273 2 s3d3"
        self.assertEqual(self.test_instance.from_unicode(s), s)
        self.assertIsNone(self.test_instance.from_unicode("null"))

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(None))
        self.assertTrue(self.test_instance.validate('string'))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate(4))


class HostnameTestCase(base.BaseTestCase):

    def setUp(self):
        super(HostnameTestCase, self).setUp()

        self.test_instance = types.Hostname()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate('ns1.ra.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('ns1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('n_s1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('n-1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('restalchemy.com'))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate('x.y.z'))
        self.assertFalse(self.test_instance.validate('restalchemy.com.'))
        self.assertFalse(self.test_instance.validate('restalchemy.com.55'))
        self.assertFalse(self.test_instance.validate('-1.55.restalchemy.com'))
        self.assertFalse(self.test_instance.validate('_s1.55.restalchemy.com'))
        self.assertFalse(self.test_instance.validate('.restalchemy.com'))
        self.assertFalse(self.test_instance.validate(u'xx.москва.рф'))
        self.assertFalse(self.test_instance.validate(u'москва.рф'))


class UrlTestCase(base.BaseTestCase):

    def setUp(self):
        super(UrlTestCase, self).setUp()

        self.test_instance = types.Url()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate(
            'http://www.gmail.com'))
        self.assertTrue(self.test_instance.validate(
            'https://www.gmail.com/test'))

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate('x.y.z'))
        self.assertFalse(self.test_instance.validate(532))
        self.assertFalse(self.test_instance.validate('google.com.55'))


class SoftSchemeDictTestCase(base.BaseTestCase):

    def setUp(self):
        super(SoftSchemeDictTestCase, self).setUp()

        self.scheme = {
            'sub_str': types.String(),
            'sub_int': types.Integer(),
        }
        self.correct_simple_data = {
            'sub_str': 'substr',
            'sub_int': 7,
        }
        self.test_instance = types.SoftSchemeDict(scheme=self.scheme)

    def test_from_simple_type(self):
        self._test_from_simple_type(
            self.test_instance, self.scheme, self.correct_simple_data
        )

    def test_from_simple_type_keys_missing(self):
        data = self.correct_simple_data.copy()
        del data[next(iter(data.keys()))]

        self._test_from_simple_type(self.test_instance, self.scheme, data)

    def test_from_simple_type_empty_data(self):
        self._test_from_simple_type(self.test_instance, self.scheme, {})

    def test_from_simple_type_extra_keys(self):
        data = self.correct_simple_data.copy()
        data['some_key'] = object()

        self.assertRaises(
            KeyError,
            self._test_from_simple_type,
            self.test_instance,
            self.scheme,
            data,
        )

    def test_validate_correct_value(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)

        self.assertTrue(
            self._test_validate(self.test_instance, self.scheme, value)
        )

    def test_validate_incorrect_value(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)
        value[next(iter(value.keys()))] = object()

        self.assertFalse(
            self._test_validate(self.test_instance, self.scheme, value)
        )

    def test_validate_keys_missing(self):
        data = self.correct_simple_data.copy()
        del data[next(iter(data.keys()))]
        value = self.test_instance.from_simple_type(data)

        self.assertTrue(
            self._test_validate(self.test_instance, self.scheme, value)
        )

    def test_validate_soft_scheme_dict_empty(self):
        value = self.test_instance.from_simple_type({})

        self.assertTrue(
            self._test_validate(self.test_instance, self.scheme, value)
        )

    def test_validate_wrong_key(self):
        value = self.test_instance.from_simple_type(self.correct_simple_data)
        first_key = next(iter(value.keys()))
        value[first_key + 'something'] = value.pop(first_key)

        self.assertFalse(
            self._test_validate(self.test_instance, self.scheme, value)
        )

    def _test_from_simple_type(self, typ_obj, scheme, data):
        arg_data = data.copy()

        with mock.patch.object(
            types.Dict, 'from_simple_type', return_value=arg_data
        ) as dict_from_simple_type:
            val = typ_obj.from_simple_type(arg_data)

        self.assertEqual(data, arg_data)
        dict_from_simple_type.assert_called_once_with(arg_data)
        self.assertEqual(data, val)
        for key, data_item in data.items():
            self.assertIsInstance(val[key], type(data_item))

    def _test_validate(self, typ_obj, scheme, value):
        arg_value = value.copy()

        with mock.patch.object(
            types.Dict, 'validate', return_value=True
        ) as dict_validate:
            ret = typ_obj.validate(arg_value)

        self.assertEqual(value, arg_value)
        dict_validate.assert_called_once_with(arg_value)

        return ret


class SchemaDictTestCase(base.BaseTestCase):
    def test_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        self.assertEqual(
            test_instance.from_simple_type({}),
            {
                "sub_str": None,
                "sub_int": None,
            },
        )

    def test_to_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        self.assertEqual(
            test_instance.to_simple_type(
                {
                    "sub_str": None,
                    "sub_int": None,
                }
            ),
            {
                "sub_str": None,
                "sub_int": None,
            },
        )

    def test_validate_wrong(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type({})
        self.assertFalse(test_instance.validate(value))

    def test_validate_correct_value_allow_none(self):
        scheme = {
            "sub_str": types.AllowNone(types.String()),
            "sub_int": types.AllowNone(types.Integer()),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type({})
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_correct_value(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
                "sub_str": "Test",
            }
        )
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_incorrect_value(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": "Test",
                "sub_str": "Test",
            }
        )
        self.assertFalse(test_instance.validate(value))

    def test_validate_with_wrong_key(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)

        self.assertFalse(
            test_instance.validate(
                {"sub_int": 1, "sub_str": "Test", "sub_wrong": "wrong_key"}
            )
        )

    def test_validate_with_wrong_key_after_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
                "sub_str": "Test",
                "sub_wrong": "wrong_key",
            }
        )
        self.assertTrue(test_instance.validate(value))

    def test_validate_with_key_missing_after_from_simple_type(self):
        scheme = {
            "sub_str": types.String(),
            "sub_int": types.Integer(),
        }

        test_instance = types.SchemeDict(scheme=scheme)
        value = test_instance.from_simple_type(
            {
                "sub_int": 1,
            }
        )
        self.assertFalse(test_instance.validate(value))



# coding=utf-8
#
#    Copyright 2021 George Melikov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from restalchemy.dm import types_network


class RecordNameTestCase(unittest.TestCase):

    def setUp(self):
        super(RecordNameTestCase, self).setUp()
        self.test_instance = types_network.RecordName()

    def test_validate_correct_value(self):
        self.assertTrue(self.test_instance.validate('ns1.ra.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('ns1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('n_s1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('n-1.55.restalchemy.com'))
        self.assertTrue(self.test_instance.validate('restalchemy.com'))
        self.assertTrue(self.test_instance.validate('a.b.c.d.1.2.3'))
        self.assertTrue(self.test_instance.validate('qa-auto-dnsidxbs'))

    def test_from_simple_type(self):
        self.assertEqual(self.test_instance.from_simple_type('.x.'), '.x')
        self.assertEqual(self.test_instance.from_simple_type('.x.x.'), '.x.x')
        self.assertEqual(self.test_instance.from_simple_type('@'), '')

    def test_to_simple_type(self):
        self.assertEqual(self.test_instance.to_simple_type(''), '@')
        self.assertEqual(self.test_instance.to_simple_type('xxx'), 'xxx')

    def test_validate_incorrect_value(self):
        self.assertFalse(self.test_instance.validate('a..b.s'))
        self.assertFalse(self.test_instance.validate('.a.b.s'))
        self.assertFalse(self.test_instance.validate('a.b.s..'))
        self.assertFalse(self.test_instance.validate(u'my.москва.рф'))
        self.assertFalse(self.test_instance.validate(u'москва.рф'))
        self.assertFalse(self.test_instance.validate(u'ee.ёёё.ЕЁ'))


class SrvNameTest(unittest.TestCase):
    def setUp(self):
        self.srv_record = types_network.SrvName()

    def test_validate(self):
        self.assertTrue(self.srv_record.validate("_sip._tcp.ra.ru"))
        self.assertTrue(self.srv_record.validate("_sip._tcp.tk"))
        self.assertTrue(self.srv_record.validate("_sip._tcp"))

        self.assertFalse(self.srv_record.validate("sip._tcp.ra.ru"))
        self.assertFalse(self.srv_record.validate("_sip.tcp.ra.ru"))
        self.assertFalse(self.srv_record.validate("hcb.jyg"))


class HostnameTest(unittest.TestCase):
    def setUp(self):
        super(HostnameTest, self).setUp()
        self.fqdn = types_network.Hostname()
        self.fqdn_2level = types_network.Hostname(min_levels=2)
        self.fqdn_with_leading_underscore = types_network.Hostname(
            allow_leading_underscore=True)
        self.fqdn_with_middle_underscore = types_network.Hostname(
            allow_middle_underscore=True)

    def test_validate(self):
        data = [
            'first-level',
            'test-me.me',
            'fe.fe',
            'aa',
            'a.bc',
            '1.2.3.4.com',
            'xn--kxae4bafwg.xn--pxaix.gr',
            'a23456789-123456789-123456789-123456789-123456789-123456789-123.'
            'b23.com',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a234567.com',
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'com',
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn.validate(fqdn), fqdn)

    def test_validate_underscore(self):
        data = [
            '_acme-challenge.example.com',
            '_acme-challenge'
        ]

        for fqdn in data:
            self.assertTrue(
                self.fqdn_with_leading_underscore.validate(fqdn),
                fqdn)

    def test_validate_underscore_negative(self):
        data = [
            '_acme-challenge._example.com'
        ]

        for fqdn in data:
            self.assertFalse(
                self.fqdn_with_leading_underscore.validate(fqdn),
                fqdn)

    def test_validate_middle_underscore(self):
        data = [
            'abc_def.example.com'
        ]

        for fqdn in data:
            self.assertTrue(
                self.fqdn_with_middle_underscore.validate(fqdn),
                fqdn)

    def test_validate_middle_underscore_negative(self):
        data = [
            'www_.example.com',
            '_www.example.com',
            'www.example._com',
            'www.example.com_',
            'www.example.co_m',
            'news.ae_ro',
        ]

        for fqdn in data:
            self.assertFalse(
                self.fqdn_with_middle_underscore.validate(fqdn),
                fqdn)

    def test_validate_negative(self):
        data = [
            'тест.ру',
            '-fe.fe',
            'fe-.fe',
            '_fe.fe',
            'f_e.fe',
            'fe_.fe',
            'a..bc',
            'ec2-35-160-210-253.us-west-2-.compute.amazonaws.com',
            '-ec2_35$160%210-253.us-west-2-.compute.amazonaws.com',
            'ec2-35-160-210-253.us-west-2-.compute.amazonaws.com',
            'a23456789-123456789-123456789-123456789-123456789-123456789-1234.'
            'b23.com',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.com',
            'mx.gmail.com.',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.com',
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn.validate(fqdn), fqdn)

    def test_min_levels(self):
        data = [
            'test-me.me',
            'fe.fe',
            'a.bc',
            '1.2.3.4.com',
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn_2level.validate(fqdn), fqdn)

    def test_min_levels_negative(self):
        data = [
            'first-level',
            'aa',
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn_2level.validate(fqdn), fqdn)


class FQDNTest(unittest.TestCase):
    def setUp(self):
        super(FQDNTest, self).setUp()
        self.fqdn = types_network.FQDN()
        self.fqdn_2level = types_network.FQDN(min_levels=2)

    def test_validate(self):
        data = [
            'first_level.',
            'test_me.me.',
            'fe.fe.',
            'aa.',
            'a.bc.',
            '1.2.3.4.com.',
            'xn--kxae4bafwg.xn--pxaix.gr.',
            'a23456789-123456789-123456789-123456789-123456789-123456789-123.'
            'b23.com.',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a234567.com.',
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcde.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijk.'
            'com.',
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn.validate(fqdn), fqdn)

    def test_validate_negative(self):
        data = [
            'тест.ру.',
            '-fe.fe.',
            'a..bc.',
            'ec2-35-160-210-253.us-west-2-.compute.amazonaws.com.',
            '-ec2_35$160%210-253.us-west-2-.compute.amazonaws.com.',
            'ec2-35-160-210-253.us-west-2-.compute.amazonaws.com.',
            'a23456789-123456789-123456789-123456789-123456789-123456789-1234.'
            'b23.com.',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.com.',
            'mx.gmail.com',
            'a23456789-a23456789-a234567890.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.a23456789.a23456789.a23456789.a23456789.a23456789.'
            'a23456789.com.',
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn.validate(fqdn), fqdn)

    def test_min_levels(self):
        data = [
            'test_me.me.',
            'fe.fe.',
            'a.bc.',
            '1.2.3.4.com.',
        ]
        for fqdn in data:
            self.assertTrue(self.fqdn_2level.validate(fqdn), fqdn)

    def test_min_levels_negative(self):
        data = [
            'first_level.',
            'aa.',
        ]
        for fqdn in data:
            self.assertFalse(self.fqdn_2level.validate(fqdn), fqdn)



# coding=utf-8
#
# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import datetime
import json
import re
import sys
import time
import uuid

import six

if six.PY2:
    # http://bugs.python.org/issue7980
    datetime.datetime.strptime("", "")


INFINITY = float("inf")
INFINITI = INFINITY  # TODO(d.burmistrov): remove this hack
UUID_RE_TEMPLATE = r"[a-f0-9]{8,8}-([a-f0-9]{4,4}-){3,3}[a-f0-9]{12,12}"

# Copy-paste from validators library because RA must support python 2.7
# and support cyrillic domain names. The validators library is located:
# https://github.com/kvesteri/validators/blob/master/validators/domain.py#L5
# the regexp has issue https://github.com/kvesteri/validators/issues/185
HOSTNAME_RE_TEMPLATE = (
    # First character of the domain
    u"^(?:[a-zA-Z0-9]"
    # Sub domain + hostname
    u"(?:[a-zA-Z0-9-_]{0,61}[A-Za-z0-9])?\.)"  # noqa
    # First 61 characters of the gTLD
    u"+[A-Za-z0-9][A-Za-z0-9-_]{0,61}"
    # Last character of the gTLD
    u"[A-Za-z]$"
)
KWARGS_OPENAPI_MAP = {
    "read_only": "readOnly",
    "default": "default",
    "example": "example",
}
MYSQL_DATETIME_FMT = "%Y-%m-%d %H:%M:%S.%f"
DEFAULT_DATE = datetime.datetime.strptime("2006-01-02 15:04:05.000576",
                                          MYSQL_DATETIME_FMT)
# RFC3339
# python's datetime doesn't support nanosecond precision
# + now without timezone, example "2006-01-02T15:04:05.999999999Z07:00"
OPENAPI_DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"


def build_prop_kwargs(kwargs):
    result = {}
    for k, v in KWARGS_OPENAPI_MAP.items():
        if k in kwargs.keys():
            value = kwargs[k]() if callable(kwargs[k]) else kwargs[k]
            if isinstance(value, (UUID, uuid.UUID)):
                # No default value for uuid
                continue
            elif isinstance(value, datetime.datetime):
                value = DEFAULT_DATE.strftime(OPENAPI_DATETIME_FMT)
            elif hasattr(value, '__dict__'):
                value = {
                    name: prop.get_property_type().to_simple_type(value[name])
                    for name, prop in value.properties.properties.items()
                }

            result[v] = value
    return result


@six.add_metaclass(abc.ABCMeta)
class BaseType(object):

    def __init__(self,
                 openapi_type="object",
                 openapi_format=None):
        super(BaseType, self).__init__()
        self._openapi_type = openapi_type
        self._openapi_format = openapi_format

    @abc.abstractmethod
    def validate(self, value):
        pass

    @abc.abstractmethod
    def to_simple_type(self, value):
        pass

    def dump_value(self, value):
        return self.to_simple_type(value)

    @abc.abstractmethod
    def from_simple_type(self, value):
        pass

    @abc.abstractmethod
    def from_unicode(self, value):
        pass

    def __repr__(self):
        return self.__class__.__name__

    @property
    def ra_type(self):
        return self.__class__

    @property
    def openapi_type(self):
        return self._openapi_type

    @property
    def openapi_format(self):
        return self._openapi_format

    def to_openapi_spec(self,
                        prop_kwargs):
        spec = {"type": self._openapi_type}
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class BasePythonType(BaseType):

    def __init__(self, python_type, **kwargs):
        super(BasePythonType, self).__init__(**kwargs)
        self._python_type = python_type

    def validate(self, value):
        return isinstance(value, self._python_type)

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return self._python_type(value)


class Boolean(BasePythonType):

    def __init__(self):
        super(Boolean, self).__init__(bool,
                                      openapi_type="boolean")

    def from_simple_type(self, value):
        return bool(value)

    def from_unicode(self, value):
        return value.lower() in ["yes", "true", "1"]


class String(BasePythonType):

    def __init__(self, min_length=0, max_length=six.MAXSIZE):
        super(String, self).__init__(six.string_types,
                                     openapi_type="string")
        self.min_length = int(min_length)
        self.max_length = int(max_length)

    def validate(self, value):
        result = super(String, self).validate(value)
        return result and self.min_length <= len(str(value)) <= self.max_length

    def from_unicode(self, value):
        return six.text_type(value)

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "minLength": self.min_length,
            "maxLength": self.max_length,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Integer(BasePythonType):

    def __init__(self, min_value=-INFINITY, max_value=INFINITY):
        super(Integer, self).__init__(six.integer_types,
                                      openapi_type="integer")
        self.min_value = (
            min_value if min_value == -INFINITY else int(min_value))
        self.max_value = max_value if max_value == INFINITY else int(max_value)

    def validate(self, value):
        result = super(Integer, self).validate(value)
        return result and self.min_value <= value <= self.max_value

    def from_unicode(self, value):
        return int(value)

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITI:
            if six.PY2:
                return sys.maxint
            else:
                return sys.maxsize
        else:
            return self.max_value

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITI:
            if six.PY2:
                return -sys.maxint
            else:
                return -sys.maxsize
        else:
            return self.min_value

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Int8(Integer):
    def __init__(self):
        super(Int8, self).__init__(
            min_value=0,
            max_value=2 ** 8 - 1)


class Float(BasePythonType):

    def __init__(self, min_value=-INFINITY, max_value=INFINITY):
        super(Float, self).__init__(float,
                                    openapi_type="number",
                                    openapi_format="float")
        self.min_value = (
            min_value if min_value == -INFINITY else float(min_value))
        self.max_value = max_value if max_value == INFINITY else float(
            max_value)

    def validate(self, value):
        result = super(Float, self).validate(value)
        return result and self.min_value <= value <= self.max_value

    @property
    def max_openapi_value(self):
        if self.max_value == INFINITI:
            if six.PY2:
                return sys.maxint
            else:
                return sys.float_info.max
        else:
            return self.max_value

    @property
    def min_openapi_value(self):
        if self.min_value == -INFINITI:
            if six.PY2:
                return -sys.maxint
            else:
                return -sys.float_info.max
        else:
            return self.min_value

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "format": self.openapi_format,
            "minimum": self.min_openapi_value,
            "maximum": self.max_openapi_value,
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class UUID(BaseType):

    def __init__(self):
        super(UUID, self).__init__(openapi_type="string",
                                   openapi_format="uuid")

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return uuid.UUID(value)

    def validate(self, value):
        return isinstance(value, uuid.UUID)

    def from_unicode(self, value):
        return uuid.UUID(value)


class ComplexPythonType(BasePythonType):

    _TYPE_ERROR_MSG = "Can't convert '%s' with type '%s' into %s"

    def _raise_on_invalid_type(self, value):
        if not isinstance(value, self._python_type):
            raise TypeError(self._TYPE_ERROR_MSG
                            % (value, type(value), self._python_type))

    def from_simple_type(self, value):
        self._raise_on_invalid_type(value)
        return value

    def from_unicode(self, value):
        result = None
        try:
            result = json.loads(value)
        except (TypeError, ValueError):
            pass
        self._raise_on_invalid_type(value)
        return self.from_simple_type(result)


class List(ComplexPythonType):

    def __init__(self):
        super(List, self).__init__(list,
                                   openapi_type="array",
                                   openapi_format="string")

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "format": self.openapi_format,
            "items": {"type": "string"}
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class TypedList(List):

    def __init__(self, nested_type):
        super(TypedList, self).__init__()
        if not isinstance(nested_type, BaseType):
            raise TypeError("Nested type '%s' is not inherited from %s"
                            % (nested_type, BaseType))
        self._nested_type = nested_type

    def validate(self, value):
        return (super(TypedList, self).validate(value)
                and all(self._nested_type.validate(item) for item in value))

    def to_simple_type(self, value):
        return [self._nested_type.to_simple_type(e) for e in value]

    def from_simple_type(self, value):
        return [self._nested_type.from_simple_type(e) for e in value]

    def from_unicode(self, value):
        if not isinstance(value, six.string_types):
            raise TypeError("Value must be six.string_types, not %s",
                            type(value))

        value = self._nested_type.from_unicode(value)
        return [value]

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "items": self._nested_type.to_openapi_spec({})
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class Dict(ComplexPythonType):

    def __init__(self):
        super(Dict, self).__init__(dict)

    def validate(self, value):
        return (super(Dict, self).validate(value)
                and all(isinstance(k, six.string_types) for k in value))

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": "object",
            "additionalProperties": {"oneOf": [{"type": "string"},
                                               {"type": "integer"},
                                               {"type": "boolean"},
                                               {"type": "object"},
                                               {"type": "array"}]}
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


def _validate_scheme(scheme):
    non_string_keys = []
    invalid_types = []

    for key, val in six.iteritems(scheme):
        if not isinstance(key, six.string_types):
            non_string_keys.append(key)
        if not isinstance(val, BaseType):
            invalid_types.append(val)

    if non_string_keys:
        raise ValueError("Scheme keys %r are not strings" % non_string_keys)
    if invalid_types:
        raise ValueError(
            "Scheme values %r are not %s instances" % (invalid_types, BaseType)
        )


# TODO(d.burmistrov): we have to make this group of Dict Schemers:
#   - ExactSchemaDict - data must follow schema in every detail
#   - PartialSchemaDict - data must be within schema definition (some keys
#                         may be missing)
#   - ExtraSchemaDict - data must follow schema but may have extra keys
#   - not sure about this option: there may be extra keys (not defined
#     in schema) and some schema keys may be missing, but all defined keys
#     matching schema must be valid due to schema

class SoftSchemeDict(Dict):

    def __init__(self, scheme):
        super(SoftSchemeDict, self).__init__()
        _validate_scheme(scheme)
        self._scheme = scheme

    def validate(self, value):
        return (super(SoftSchemeDict, self).validate(value)
                and set(value.keys()).issubset(set(self._scheme.keys()))
                and all(self._scheme[k].validate(v)
                        for k, v in six.iteritems(value)))

    def to_simple_type(self, value):
        return {k: self._scheme[k].to_simple_type(v)
                for k, v in six.iteritems(value)}

    def from_simple_type(self, value):
        value = super(SoftSchemeDict, self).from_simple_type(value)
        return {k: self._scheme[k].from_simple_type(v)
                for k, v in six.iteritems(value)}

    def to_openapi_spec(self,
                        prop_kwargs):
        spec = {
            "type": "object",
            "properties": {}
        }
        for k, v in self.__scheme__.items():
            spec["properties"][k] = v.to_openapi_spec(prop_kwargs)
        return spec


class SchemeDict(Dict):

    def __init__(self, scheme):
        super(SchemeDict, self).__init__()
        _validate_scheme(scheme)
        self._scheme = scheme

    def validate(self, value):
        return (super(SchemeDict, self).validate(value)
                and set(value.keys()) == set(self._scheme.keys())
                and all(scheme.validate(value[key])
                        for key, scheme in six.iteritems(self._scheme)))

    def to_simple_type(self, value):
        return {key: scheme.to_simple_type(value[key])
                for key, scheme in self._scheme.items()}

    def from_simple_type(self, value):
        value = super(SchemeDict, self).from_simple_type(value)
        return {key: scheme.from_simple_type(value.get(key, None))
                for key, scheme in six.iteritems(self._scheme)}

    def to_openapi_spec(self,
                        prop_kwargs):
        spec = {
            "type": "object",
            "properties": {}
        }
        for k, v in self.__scheme__.items():
            spec["properties"][k] = v.to_openapi_spec(prop_kwargs)
        return spec


class TypedDict(Dict):

    def __init__(self, nested_type):
        super(TypedDict, self).__init__()
        if not isinstance(nested_type, BaseType):
            raise TypeError("Nested type '%s' is not inherited from %s"
                            % (nested_type, BaseType))
        self._nested_type = nested_type

    def validate(self, value):
        return (super(TypedDict, self).validate(value)
                and all(self._nested_type.validate(element)
                        for element in six.itervalues(value)))

    def to_simple_type(self, value):
        return {k: self._nested_type.to_simple_type(v)
                for k, v in six.iteritems(value)}

    def from_simple_type(self, value):
        value = super(TypedDict, self).from_simple_type(value)
        return {k: self._nested_type.from_simple_type(v)
                for k, v in six.iteritems(value)}

    def to_openapi_spec(self,
                        prop_kwargs):
        spec = {
            "type": self._openapi_type,
            "additionalProperties": self._nested_type.to_openapi_spec({})
        }
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class UTCDateTime(BasePythonType):

    def __init__(self):
        super(UTCDateTime, self).__init__(python_type=datetime.datetime,
                                          openapi_type="string",
                                          openapi_format="date-time")

    def validate(self, value):
        return isinstance(value, datetime.datetime) and value.tzinfo is None

    def to_simple_type(self, value):
        return value.strftime(MYSQL_DATETIME_FMT)

    def dump_value(self, value):
        # Converting value in api response
        return value.strftime(OPENAPI_DATETIME_FMT)

    def from_simple_type(self, value):
        if isinstance(value, datetime.datetime):
            return value
        try:
            return datetime.datetime.strptime(value, MYSQL_DATETIME_FMT)
        except ValueError:
            # Used in cases, than we manually convert openapi string in
            # http response to ra type in model
            return datetime.datetime.strptime(
                datetime.datetime.strptime(
                    value, OPENAPI_DATETIME_FMT).strftime(MYSQL_DATETIME_FMT),
                MYSQL_DATETIME_FMT
            )

    def from_unicode(self, value):
        return self.from_simple_type(value)

    def to_openapi_spec(self,
                        prop_kwargs):
        spec = {
            "type": self._openapi_type,
            # https://github.com/ogen-go/ogen/blob/main/_testdata/positive/time_extension.yml#L29
            "x-ogen-time-format": self.dump_value(DEFAULT_DATE)
        }
        if self._openapi_format is not None:
            spec["format"] = self._openapi_format
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        if "default" in prop_kwargs and (
                self.validate(prop_kwargs["default"])
                or callable(prop_kwargs["default"])):
            spec["default"] = self.dump_value(prop_kwargs["default"]())
        return spec


class DateTime(BasePythonType):

    def __init__(self):
        super(DateTime, self).__init__(python_type=datetime.datetime)

    def to_simple_type(self, value):
        return int(time.mktime(value.utctimetuple()))

    def from_simple_type(self, value):
        return datetime.datetime.utcfromtimestamp(value)

    def from_unicode(self, value):
        return self.from_simple_type(value)


class Enum(BaseType):

    def __init__(self, enum_values):
        super(Enum, self).__init__(openapi_type="string")
        self._enums_values = copy.deepcopy(enum_values)

    @property
    def values(self):
        return self._enums_values

    def validate(self, value):
        return value in self._enums_values

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        for enum_value in self._enums_values:
            if value == six.text_type(enum_value):
                return enum_value
        raise TypeError("Can't convert '%s' to enum type."
                        " Allowed values are %s"
                        % (value, self._enums_values))

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "enum": sorted(list(self.values)),
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class BaseRegExpType(BaseType):
    """BaseCompiledRegExpTypeFromAttr is preferred to be used"""

    def __init__(self, pattern,
                 openapi_type="string",
                 **kwargs):
        super(BaseRegExpType, self).__init__(openapi_type, **kwargs)
        self._pattern = re.compile(pattern)

    def validate(self, value):
        try:
            return self._pattern.match(value) is not None
        except TypeError:
            return False

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return value

    @property
    def pattern_openapi(self):
        return self._pattern.pattern

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "pattern": self.pattern_openapi
        }
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class BaseCompiledRegExpType(BaseRegExpType):
    def __init__(self, pattern, **kwargs):
        if "openapi_type" not in kwargs:
            kwargs["openapi_type"] = "string"
        super(BaseRegExpType, self).__init__(**kwargs)
        self._pattern = pattern


class BaseCompiledRegExpTypeFromAttr(BaseCompiledRegExpType):
    def __init__(self, **kwargs):
        super(BaseCompiledRegExpTypeFromAttr, self).__init__(
            pattern=self.pattern, **kwargs)


class Uri(BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^(/[A-Za-z0-9\-_]*)*/%s$" % UUID_RE_TEMPLATE)

    def __init__(self):
        super(Uri, self).__init__(
            openapi_type="string",
            openapi_format="uri",
        )


class Mac(BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^([0-9a-fA-F]{2,2}:){5,5}[0-9a-fA-F]{2,2}$")

    def __init__(self):
        super(Mac, self).__init__(
            openapi_type="string",
            openapi_format="mac",
        )


class Hostname(BaseCompiledRegExpTypeFromAttr):
    """DEPRECATED! Use types from types_network module"""
    pattern = re.compile(HOSTNAME_RE_TEMPLATE)

    def __init__(self):
        super(Hostname, self).__init__(
            openapi_type="hostname",
        )


class Url(BaseCompiledRegExpTypeFromAttr):
    """

    django url validation regex
    (https://github.com/django/django/blob/stable/1.3.x/django/core/validators.py#L45):
    """

    pattern = re.compile(
        r"^(?:http|ftp)s?://"
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # noqa
        r"localhost|"
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
        r"(?::\d+)?"
        r"(?:/?|[/?]\S+)$", re.IGNORECASE)


class AllowNone(BaseType):

    def __init__(self, nested_type):
        super(AllowNone, self).__init__(openapi_type="string")
        self._nested_type = nested_type

    @property
    def nested_type(self):
        return self._nested_type

    def validate(self, value):
        return value is None or self._nested_type.validate(value)

    def to_simple_type(self, value):
        return None if value is None else self._nested_type.to_simple_type(
            value)

    def from_simple_type(self, value):
        return None if value is None else self._nested_type.from_simple_type(
            value)

    def from_unicode(self, value):
        if value == "null":
            return None
        else:
            return self._nested_type.from_unicode(value)

    def to_openapi_spec(self, prop_kwargs):
        if "default" in prop_kwargs and prop_kwargs["default"] is None:
            del prop_kwargs["default"]
        spec = self._nested_type.to_openapi_spec(prop_kwargs)
        spec["nullable"] = True
        return spec



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# HTTP methods
GET = 'GET'
PUT = 'PUT'
POST = 'POST'

# RA methods
FILTER = 'FILTER'
CREATE = 'CREATE'
UPDATE = 'UPDATE'
DELETE = 'DELETE'

# RA action methods
ACTION_GET = 'ACTION_GET'
ACTION_POST = 'ACTION_POST'
ACTION_PUT = 'ACTION_PUT'

# all RA methods
ALL = 'ALL'

ALL_RA_METHODS = [
    GET,              # Controller.get
    FILTER,           # Controller.filter
    CREATE,           # Controller.create
    UPDATE,           # Controller.update
    DELETE,           # Controller.delete
    ACTION_GET,       # Action.get
    ACTION_POST,      # Action.post
    ACTION_PUT,       # Action.put
    ALL               # All RA methods + actions
]


HTTP_TO_RA_COLLECTION_METHODS = {GET: FILTER, POST: CREATE}
HTTP_TO_RA_RESOURCE_METHODS = {GET: GET, PUT: UPDATE, DELETE: DELETE}

CONTENT_TYPE_APPLICATION_JSON = DEFAULT_CONTENT_TYPE = 'application/json'



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2022 George Melikov
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
import webob

from restalchemy.api import constants
from restalchemy.api import contexts
from restalchemy.api import field_permissions
from restalchemy.api import resources
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types


class FakeModel(models.CustomPropertiesMixin, models.ModelWithUUID):
    _private_field = properties.property(types.Integer())
    standard_field1 = properties.property(types.Integer(), required=True)
    standard_field2 = properties.property(types.Integer())
    standard_field3 = properties.property(types.Integer())
    standard_field4 = properties.property(types.Integer(), required=True)
    standard_field5 = properties.property(types.Integer())


class FakeMemberContext(object):
    roles = ['_member_']


class FakeAdminContext(object):
    roles = ['admin']


class FakeEmptyContext(object):
    roles = []


class ResourceByRAModelFieldsPermissions(unittest.TestCase):

    def setUp(self):
        super(ResourceByRAModelFieldsPermissions,
              self).setUp()

        self.resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=field_permissions.FieldsPermissionsByRole(
                default=field_permissions.UniversalPermissions(
                    permission=field_permissions.Permissions.HIDDEN),
                admin=field_permissions.FieldsPermissions(
                    {
                        'standard_field1': {
                            constants.ALL: field_permissions.Permissions.HIDDEN
                        },
                        'standard_field2': {
                            constants.CREATE: (
                                field_permissions.Permissions.HIDDEN)
                        },
                        'standard_field3': {
                            constants.GET: field_permissions.Permissions.HIDDEN
                        },
                        'standard_field5': {
                            constants.FILTER: (
                                field_permissions.Permissions.HIDDEN
                            ),
                            constants.DELETE: (
                                field_permissions.Permissions.HIDDEN
                            )
                        },
                    }
                ),
                _member_=field_permissions.FieldsPermissions(
                    {
                        'standard_field1': {
                            constants.CREATE: (
                                field_permissions.Permissions.HIDDEN),
                            constants.ALL: field_permissions.Permissions.RO
                        },
                        'standard_field2': {
                            constants.GET: field_permissions.Permissions.RO,
                            constants.FILTER: field_permissions.Permissions.RO,
                            constants.ALL: field_permissions.Permissions.HIDDEN
                        },
                        'standard_field4': {
                            constants.GET: field_permissions.Permissions.RO,
                            constants.UPDATE: field_permissions.Permissions.RO,
                            constants.ALL: field_permissions.Permissions.HIDDEN
                        },
                        'standard_field5': {
                            constants.UPDATE: (
                                field_permissions.Permissions.HIDDEN
                            ),
                            constants.DELETE: (
                                field_permissions.Permissions.HIDDEN
                            )
                        },
                    }
                ),
            )
        )

        self._request = webob.Request.blank('/some-uri')
        self._request.api_context = contexts.RequestContext(self._request)

    def tearDown(self):
        super(ResourceByRAModelFieldsPermissions,
              self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._request

    def _test_fields_is_shown(self, expected_fields):
        result = [name
                  for name, prop in self.resource.get_fields_by_request(
                      self._request
                  )
                  if prop.is_public()
                  and not self.resource._fields_permissions.is_hidden(
                      name,
                      self._request
                  )
                  ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_fields_is_hidden(self, expected_fields):
        result = [name
                  for name, prop in self.resource.get_fields_by_request(
                      self._request
                  )
                  if prop.is_public()
                  and self.resource._fields_permissions.is_hidden(
                      name,
                      self._request
                  )
                  ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_fields_is_readonly(self, expected_fields):
        result = [name
                  for name, prop in self.resource.get_fields_by_request(
                      self._request
                  )
                  if prop.is_public()
                  and self.resource._fields_permissions.is_readonly(
                      name,
                      self._request
                  )
                  ]

        self.assertEqual(sorted(expected_fields), sorted(result))

    def _test_default(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=field_permissions.FieldsPermissionsByRole(
                default=field_permissions.UniversalPermissions(
                    permission=field_permissions.Permissions.HIDDEN),
                admin=field_permissions.FieldsPermissions(
                    {
                        'standard_field1': {
                            constants.ALL: field_permissions.Permissions.HIDDEN
                        },
                    },
                ),
            )
        )

        self.assertFalse(resource._fields_permissions.is_readonly(
                         'standard_field2',
                         self._request
                         ))

    def _test_default_custom_value(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            fields_permissions=field_permissions.FieldsPermissionsByRole(
                default=field_permissions.UniversalPermissions(
                    permission=field_permissions.Permissions.HIDDEN),
                admin=field_permissions.FieldsPermissions(
                    {
                        'standard_field1': {
                            constants.ALL: field_permissions.Permissions.HIDDEN
                        },
                    },
                    default=field_permissions.Permissions.RO
                ),
            )
        )

        self.assertTrue(resource._fields_permissions.is_readonly(
                        'standard_field2',
                        self._request
                        ))


class ResourceByRAModelFieldsPermissionsRoleAdmin(
    ResourceByRAModelFieldsPermissions):

    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsRoleAdmin,
              self).setUp()
        admin_context = FakeAdminContext()
        self._request.context = admin_context

    def test_fields_permission_role_admin_method_get(self):
        self._request.api_context.set_active_method(constants.GET)

        expected_shown_fields = [
            'standard_field2',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1',
            'standard_field3'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1',
            'standard_field5'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)

        expected_shown_fields = [
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1',
            'standard_field2',
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1',
            'standard_field5'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_admin_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)

        expected_shown_fields = [
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field1'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)


class ResourceByRAModelFieldsPermissionsNoRole(
    ResourceByRAModelFieldsPermissions):

    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsNoRole,
              self).setUp()
        empty_context = FakeEmptyContext()
        self._request.context = empty_context

        self.expected_shown_fields = []
        self.expected_hidden_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]

    def test_fields_permission_no_role_method_get(self):
        self._request.api_context.set_active_method(constants.GET)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)

    def test_fields_permission_no_role_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)

        self._test_fields_is_shown(self.expected_shown_fields)
        self._test_fields_is_hidden(self.expected_hidden_fields)
        self._test_fields_is_readonly(self.expected_hidden_fields)


class ResourceByRAModelFieldsPermissionsRoleMember(
    ResourceByRAModelFieldsPermissions):

    def setUp(self):
        super(ResourceByRAModelFieldsPermissionsRoleMember,
              self).setUp()
        member_context = FakeMemberContext()
        self._request.context = member_context

    def test_fields_permission_role_member_method_get(self):
        self._request.api_context.set_active_method(constants.GET)
        expected_shown_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field3',
            'standard_field4',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4',
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_update(self):
        self._request.api_context.set_active_method(constants.UPDATE)
        expected_shown_fields = [
            'standard_field1',
            'standard_field3',
            'standard_field4',
            'uuid']
        expected_hidden_fields = [
            'standard_field2',
            'standard_field5'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4',
            'standard_field5'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_create(self):
        self._request.api_context.set_active_method(constants.CREATE)
        expected_shown_fields = [
            'standard_field3',
            'standard_field5',
            'uuid']
        expected_hidden_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_hidden_fields)

    def test_fields_permission_role_member_method_delete(self):
        self._request.api_context.set_active_method(constants.DELETE)
        expected_shown_fields = [
            'standard_field1',
            'standard_field3',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field2',
            'standard_field4',
            'standard_field5'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4',
            'standard_field5'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_method_filter(self):
        self._request.api_context.set_active_method(constants.FILTER)
        expected_shown_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field3',
            'standard_field5',
            'uuid'
        ]
        expected_hidden_fields = [
            'standard_field4'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_get(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)
        expected_shown_fields = [
            'standard_field1',
            'standard_field3',
            'standard_field5',
            'uuid']
        expected_hidden_fields = [
            'standard_field2',
            'standard_field4'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_post(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)
        expected_shown_fields = [
            'standard_field1',
            'standard_field3',
            'standard_field5',
            'uuid']
        expected_hidden_fields = [
            'standard_field2',
            'standard_field4'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)

    def test_fields_permission_role_member_action_put(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)
        expected_shown_fields = [
            'standard_field1',
            'standard_field3',
            'standard_field5',
            'uuid']
        expected_hidden_fields = [
            'standard_field2',
            'standard_field4'
        ]
        expected_ro_fields = [
            'standard_field1',
            'standard_field2',
            'standard_field4'
        ]

        self._test_fields_is_shown(expected_shown_fields)
        self._test_fields_is_hidden(expected_hidden_fields)
        self._test_fields_is_readonly(expected_ro_fields)



# Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
import uuid

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.storage.sql import orm


FAKE_UUID0 = uuid.UUID('00000000-0000-0000-0000-000000000000')
FAKE_UUID1 = uuid.UUID('00000000-0000-0000-0000-000000000001')
FAKE_UUID2 = uuid.UUID('00000000-0000-0000-0000-000000000002')


class SimpleModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = 'simple_table'

    field_str = properties.property(types.String(), default='FAKE_STR')
    field_int = properties.property(types.Integer(), default=1)
    field_bool = properties.property(types.Boolean(), default=True)


class ModelWithL1Relationships(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = 'model_with_l1_relationships'

    ref_l0_1 = relationships.relationship(SimpleModel, prefetch=True)
    ref_l0_2 = relationships.relationship(SimpleModel, prefetch=False)


class ModelWithL2Relationships(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = 'model_with_l2_relationships'

    ref_l1_1 = relationships.relationship(ModelWithL1Relationships,
                                          prefetch=True)
    ref_l1_2 = relationships.relationship(SimpleModel, prefetch=True)
    ref_l1_3 = relationships.relationship(ModelWithL1Relationships,
                                          prefetch=False)


# NOTE(efrolov): Sort model properties for correct ordering in asserts
for model in [SimpleModel, ModelWithL1Relationships, ModelWithL2Relationships]:
    model.properties.sort_properties()


class MySQLPrefetchQueryBuilderTestCase(unittest.TestCase):

    def setUp(self):
        super(MySQLPrefetchQueryBuilderTestCase, self).setUp()
        self.Q = q.Q
        model1 = SimpleModel(uuid=FAKE_UUID1)
        model2 = SimpleModel(uuid=FAKE_UUID2)
        self.flt = filters.AND(
            {'uuid': filters.EQ(FAKE_UUID0)},
            {'ref_l0_1': filters.LE(model1)},
            {'ref_l0_2': filters.NE(model2)},
        )

    def tearDown(self):
        super(MySQLPrefetchQueryBuilderTestCase, self).tearDown()
        del self.Q

    def test_l1_select(self):
        query = self.Q.select(ModelWithL1Relationships)

        result = query.compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`)",
            result
        )

    def test_l1_select_with_filters(self):
        query = self.Q.select(ModelWithL1Relationships).where(self.flt)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "WHERE"
            " (`t1`.`uuid` = %s AND"
            " `t1`.`ref_l0_1` <= %s AND"
            " `t1`.`ref_l0_2` <> %s)",
            result_expression
        )
        self.assertEqual(
            ['00000000-0000-0000-0000-000000000000',
             '00000000-0000-0000-0000-000000000001',
             '00000000-0000-0000-0000-000000000002'],
            result_values
        )

    def test_select_with_limit(self):
        query = self.Q.select(ModelWithL1Relationships).limit(100500)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "LIMIT 100500",
            result_expression
        )
        self.assertEqual([], result_values)

    def test_l1_select_lock_with_filters(self):
        query = self.Q.select(ModelWithL1Relationships).where(self.flt).for_()

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "WHERE"
            " (`t1`.`uuid` = %s AND"
            " `t1`.`ref_l0_1` <= %s AND"
            " `t1`.`ref_l0_2` <> %s) "
            "FOR UPDATE",
            result_expression
        )
        self.assertEqual(
            ['00000000-0000-0000-0000-000000000000',
             '00000000-0000-0000-0000-000000000001',
             '00000000-0000-0000-0000-000000000002'],
            result_values
        )

    def test_l1_select_order_by(self):
        query = self.Q.select(ModelWithL1Relationships).order_by('uuid')

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "ORDER BY"
            " `t1_uuid` ASC",
            result_expression
        )
        self.assertEqual([], result_values)

    def test_l2_select(self):
        result = self.Q.select(ModelWithL2Relationships).compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l1_3` AS `t1_ref_l1_3`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`ref_l0_2` AS `t2_ref_l0_2`,"
            " `t2`.`uuid` AS `t2_uuid`,"
            " `t3`.`field_bool` AS `t3_field_bool`,"
            " `t3`.`field_int` AS `t3_field_int`,"
            " `t3`.`field_str` AS `t3_field_str`,"
            " `t3`.`uuid` AS `t3_uuid`,"
            " `t4`.`field_bool` AS `t4_field_bool`,"
            " `t4`.`field_int` AS `t4_field_int`,"
            " `t4`.`field_str` AS `t4_field_str`,"
            " `t4`.`uuid` AS `t4_uuid` "
            "FROM"
            " `model_with_l2_relationships` AS `t1` "
            "LEFT JOIN"
            " `model_with_l1_relationships` AS `t2` "
            "ON"
            " (`t1`.`ref_l1_1` = `t2`.`uuid`) "
            "LEFT JOIN"
            " `simple_table` AS `t3` "
            "ON"
            " (`t2`.`ref_l0_1` = `t3`.`uuid`) "
            "LEFT JOIN"
            " `simple_table` AS `t4` "
            "ON"
            " (`t1`.`ref_l1_2` = `t4`.`uuid`)",
            result
        )


class MySQLResultParserTestCase(unittest.TestCase):

    def test_l1_prefetch_result_parser(self):
        row_from_db = {
            "t1_ref_l0_2": "FakeRsUUID",
            "t1_uuid": "FakeUUID0",
            "t2_field_bool": "FakeBool",
            "t2_field_int": "FakeInt",
            "t2_field_str": "FakeStr",
            "t2_uuid": "FakeUUID1",
        }
        select_clause = q.Q.select(ModelWithL1Relationships)

        result = select_clause.parse_row(row_from_db)

        self.assertEqual(
            {
                "uuid": "FakeUUID0",
                "ref_l0_1": {
                    "field_bool": "FakeBool",
                    "field_int": "FakeInt",
                    "field_str": "FakeStr",
                    "uuid": "FakeUUID1",
                },
                "ref_l0_2": "FakeRsUUID",
            },
            result,
        )



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.common import exceptions
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.tests.unit import base

FAKE_VALUE = 'FAKE_VALUE'
FAKE_VALUE2 = 'FAKE_VALUE2'
FAKE_VALUE3 = 'FAKE_VALUE3'


class PropertyTestCase(base.BaseTestCase):

    def setUp(self):
        super(PropertyTestCase, self).setUp()
        self.positive_fake_property_type = mock.MagicMock(
            **{'validate': mock.MagicMock(return_value=True),
               'spec': types.BaseType})
        self.negative_fake_property_type = mock.MagicMock(
            **{'validate': mock.MagicMock(return_value=False),
               'spec': types.BaseType})

    def _set_property_value(self, obj, value):
        obj.value = value

    def test_init_with_correct_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, value=FAKE_VALUE)

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_init_incorrect_value(self):
        self.assertRaises(exceptions.TypeError, properties.Property,
                          self.negative_fake_property_type, value=FAKE_VALUE)

    def test_init_default_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, default=FAKE_VALUE)

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_init_default_callable_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, default=lambda: FAKE_VALUE)

        self.assertEqual(property_obj.value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_init_default_value_if_property_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, default=FAKE_VALUE,
            read_only=True)

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_init_none_and_require(self):
        self.assertRaises(exceptions.PropertyRequired, properties.Property,
                          self.negative_fake_property_type, required=True,
                          value=None)

    def test_init_value_is_none_default_is_not_none_and_prop_required(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, value=None, required=True,
            default=FAKE_VALUE)

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_init_none_value(self):
        property_obj = properties.Property(
            self.negative_fake_property_type, value=None)

        self.assertEqual(property_obj._value, None)

    def test_init_value_with_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True,
            value=FAKE_VALUE2)

        self.assertEqual(property_obj._value, FAKE_VALUE2)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE2)

    def test_set_correct_value(self):
        property_obj = properties.Property(self.positive_fake_property_type)

        self.assertIsNone(self._set_property_value(property_obj, FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_set_incorect_value(self):
        property_obj = properties.Property(self.negative_fake_property_type)
        old_value = property_obj._value

        self.assertRaises(exceptions.TypeError, self._set_property_value,
                          property_obj, FAKE_VALUE)
        self.assertEqual(property_obj._value, old_value)
        self.negative_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_set_value_if_property_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True)

        self.assertRaises(exceptions.ReadOnlyProperty,
                          self._set_property_value, property_obj, FAKE_VALUE2)
        self.assertEqual(property_obj._value, None)

    def test_set_force_correct_value(self):
        property_obj = properties.Property(self.positive_fake_property_type)

        self.assertIsNone(property_obj.set_value_force(FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_set_force_incorect_value(self):
        property_obj = properties.Property(self.negative_fake_property_type)
        old_value = property_obj._value

        self.assertRaises(exceptions.TypeError, property_obj.set_value_force,
                          FAKE_VALUE)
        self.assertEqual(property_obj._value, old_value)
        self.negative_fake_property_type.validate.assert_called_with(
            FAKE_VALUE)

    def test_set_force_none_value(self):
        property_obj = properties.Property(self.positive_fake_property_type,
                                           required=True, value=FAKE_VALUE)
        old_value = property_obj._value

        self.assertRaises(exceptions.PropertyRequired,
                          property_obj.set_value_force, None)
        self.assertEqual(property_obj._value, old_value)


class PropertyCreatorTestCase(base.BaseTestCase):

    ARGS = [1, 2, 3]
    KWARGS = {'fake_key1': 'fake_value1', 'fake_key2': 'fake_value2'}

    def setUp(self):
        self.property_mock = mock.Mock(return_value=FAKE_VALUE)
        self.prop_type_mock = mock.Mock()
        self.test_instance = properties.PropertyCreator(self.property_mock,
                                                        self.prop_type_mock,
                                                        self.ARGS, self.KWARGS)

    def test_call_object(self):
        self.assertEqual(self.test_instance(FAKE_VALUE2), FAKE_VALUE)
        self.property_mock.assert_called_once_with(
            value=FAKE_VALUE2,
            property_type=self.prop_type_mock,
            *self.ARGS, **self.KWARGS)

    def test_get_property_class(self):
        self.assertEqual(self.test_instance.get_property_class(),
                         self.property_mock)


class PropertyCollectionTestCase(base.BaseTestCase):

    def setUp(self):

        self.kwargs = {
            'fake1': mock.Mock(
                **{'get_property_class.return_value': FAKE_VALUE,
                   'return_value': FAKE_VALUE2}),
            'fake2': mock.Mock(
                **{'get_property_class.return_value': FAKE_VALUE2,
                   'return_value': FAKE_VALUE})
        }
        self.test_instance = properties.PropertyCollection(**self.kwargs)

    def test_properties(self):
        self.assertEqual(self.test_instance.properties, self.kwargs)

    def test_change_properties_dict(self):
        def set_item(d, key, value):
            d[key] = value

        self.assertRaises(TypeError, set_item, self.test_instance.properties,
                          'fake1', 2)

    def test_get_item_fake1(self):
        self.assertEqual(self.test_instance['fake1'], FAKE_VALUE)
        self.assertTrue(self.kwargs['fake1'].get_property_class.called)
        self.assertFalse(self.kwargs['fake2'].get_property_class.called)

    def test_get_item_fake2(self):
        self.assertEqual(self.test_instance['fake2'], FAKE_VALUE2)
        self.assertFalse(self.kwargs['fake1'].get_property_class.called)
        self.assertTrue(self.kwargs['fake2'].get_property_class.called)

    def test_instantiate_property_fake1(self):
        self.assertEqual(self.test_instance.instantiate_property(
            name='fake1', value=FAKE_VALUE3), FAKE_VALUE2)
        self.kwargs['fake1'].assert_called_once_with(FAKE_VALUE3)
        self.assertFalse(self.kwargs['fake2'].called)

    def test_instantiate_property_fake2(self):
        self.assertEqual(self.test_instance.instantiate_property(
            name='fake2', value=FAKE_VALUE3), FAKE_VALUE)
        self.kwargs['fake2'].assert_called_once_with(FAKE_VALUE3)
        self.assertFalse(self.kwargs['fake1'].called)

    def test_concatenate_two_collections(self):
        fake_property3 = mock.Mock()
        test_instance2 = properties.PropertyCollection(fake3=fake_property3)
        new_properties = self.kwargs
        new_properties['fake3'] = fake_property3

        res = self.test_instance + test_instance2
        self.assertIsInstance(res, properties.PropertyCollection)
        self.assertNotEqual(res, self.test_instance)
        self.assertNotEqual(res, test_instance2)
        self.assertEqual(res._properties, new_properties)

    def test_concatenate_collection_and_other_object(self):

        def concatenate(a, b):
            return a + b

        self.assertRaises(TypeError, concatenate, self.test_instance, object())


class PropertyManagerTestCase(base.BaseTestCase):

    def setUp(self):
        self.collection_mock = mock.Mock(**{
            'properties.items.return_value': [
                ('fake1', 'fake1'),
                ('fake2', 'fake2')
            ],
            'instantiate_property.return_value': FAKE_VALUE})

    def test_init_manager(self):
        res = properties.PropertyManager(self.collection_mock)
        instantiate_property_calls = [
            mock.call('fake1', None),
            mock.call('fake2', None)]

        self.assertIsInstance(res, properties.PropertyManager)
        self.collection_mock.instantiate_property.assert_has_calls(
            instantiate_property_calls, any_order=True)

    def test_init_manager_with_correct_kwargs(self):
        res = properties.PropertyManager(self.collection_mock,
                                         fake1=FAKE_VALUE2,
                                         fake2=FAKE_VALUE3)
        instantiate_property_calls = [
            mock.call('fake1', FAKE_VALUE2),
            mock.call('fake2', FAKE_VALUE3)]

        self.assertIsInstance(res, properties.PropertyManager)
        self.collection_mock.instantiate_property.assert_has_calls(
            instantiate_property_calls, any_order=True)

    @base.unittest.skip("Checking of redundant data is turned off.")
    def test_init_manager_with_incorrect_kwargs(self):
        self.assertRaises(TypeError, properties.PropertyManager,
                          self.collection_mock, fake3=FAKE_VALUE3)

    def test_properties(self):
        property_manager = properties.PropertyManager(self.collection_mock)

        self.assertEqual(property_manager.properties, {'fake1': 'FAKE_VALUE',
                                                       'fake2': 'FAKE_VALUE'})

    def test_change_properties_dict(self):
        property_manager = properties.PropertyManager(self.collection_mock)

        def set_item(d, key, value):
            d[key] = value

        self.assertRaises(TypeError, set_item, property_manager.properties,
                          'fake1', 2)


@mock.patch('restalchemy.dm.properties.PropertyCreator',
            return_value=FAKE_VALUE)
class PropertyFuncTestCase(base.BaseTestCase):

    ARGS = (1, 2, 3)
    KWARGS = {'fake_key1': 'fake_value1', 'fake_key2': 'fake_value2'}

    def test_create_property(self, pc_mock):
        self.assertEqual(properties.property(*self.ARGS, **self.KWARGS),
                         FAKE_VALUE)
        pc_mock.assert_called_once_with(
            prop_class=properties.Property, prop_type=1, args=self.ARGS[1:],
            kwargs=self.KWARGS)

    def test_create_property_with_property_class(self, pc_mock):

        class LocalProperty(properties.AbstractProperty):

            @property
            def value(self):
                pass

            def set_value_force(self, value):
                pass

        self.assertEqual(properties.property(
            property_class=LocalProperty, *self.ARGS, **self.KWARGS),
            FAKE_VALUE)
        pc_mock.assert_called_once_with(
            prop_class=LocalProperty, prop_type=1, args=self.ARGS[1:],
            kwargs=self.KWARGS)

    def test_create_property_with_incorect_property_class(self, pc_mock):
        self.assertRaises(ValueError, properties.property,
                          property_class=object, *self.ARGS, **self.KWARGS)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2021 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.api.middlewares.logging import LoggingMiddleware
from restalchemy.tests.unit import base


class LoggingMiddlewareTestCase(base.BaseTestCase):

    def test_sanitize_authorization_header(self):
        request_mock = mock.MagicMock()
        request_mock.headers = {
            'fake_header': 'fake_header_value',
            'Authorization': 'basic something'}
        middlew = LoggingMiddleware('application')
        checks = {}

        def _check_sanitized_header(msg, *args, **kwargs):
            if 'API > ' in msg:  # test only request logs
                checks['check_run'] = True
                for header in args[1]:
                    if header.startswith('Authorization'):
                        if 'something' not in header:
                            checks['Authorization'] = True
                    if header.startswith('fake_header'):
                        if 'fake_header_value' in header:
                            checks['fake_header'] = True

        middlew.logger.debug = mock.Mock(side_effect=_check_sanitized_header)

        middlew.process_request(request_mock)

        self.assertTrue(checks['check_run'])
        self.assertTrue(checks['Authorization'])
        self.assertTrue(checks['fake_header'])
        self.assertDictEqual(
            {
                'fake_header': 'fake_header_value',
                'Authorization': 'basic something'
            },
            request_mock.headers
        )



#    Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#    Copyright 2021 Eugene Frolov.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json
import logging
import types

import six

from restalchemy.api import constants
from restalchemy.common import exceptions
from restalchemy.common import utils

DEFAULT_VALUE = object()
CONTENT_TYPE_APPLICATION_JSON = DEFAULT_CONTENT_TYPE = constants.CONTENT_TYPE_APPLICATION_JSON  # noqa

LOG = logging.getLogger(__name__)


def get_content_type(headers):
    return headers.get('Content-Type') or constants.DEFAULT_CONTENT_TYPE


class BaseResourcePacker(object):

    _skip_none = True

    def __init__(self, resource_type, request):
        self._rt = resource_type
        self._req = request

    def pack_resource(self, obj):
        if isinstance(obj, six.string_types + six.integer_types + (
                float, bool, type(None), list, tuple, dict)):
            return obj
        else:
            result = {}
            for name, prop in self._rt.get_fields_by_request(self._req):
                api_name = prop.api_name
                if (prop.is_public()
                    and not self._rt._fields_permissions.is_hidden(
                        self._req,
                        name)):
                    value = getattr(obj, name)
                    if value is None:
                        if not self._skip_none:
                            result[api_name] = value
                    else:
                        result[api_name] = prop.dump_value(value)

            return result

    def pack(self, obj):
        if (isinstance(obj, list)
                or isinstance(obj, types.GeneratorType)):
            return [self.pack_resource(resource) for resource in obj]
        else:
            return self.pack_resource(obj)

    @utils.raise_parse_error_on_fail
    def _parse_value(self, name, value, prop):
        return prop.parse_value(self._req, value)

    def unpack(self, value):
        if not self._rt:
            return value
        value = copy.deepcopy(value)
        result = {}
        for name, prop in self._rt.get_fields_by_request(self._req):
            api_name = prop.api_name
            prop_value = value.pop(api_name, DEFAULT_VALUE)
            if prop_value is not DEFAULT_VALUE:
                if not prop.is_public():
                    raise exceptions.ValidationPropertyPrivateError(
                        property=api_name)

                if self._rt._fields_permissions.is_readonly(self._req, name):
                    raise exceptions.FieldPermissionError(
                        field=name
                    )
                result[name] = self._parse_value(api_name, prop_value, prop)

        if len(value) > 0:
            raise TypeError("%s is not compatible with %s" % (value, self._rt))

        return result


class JSONPacker(BaseResourcePacker):

    def pack(self, obj):
        return json.dumps(super(JSONPacker, self).pack(obj))

    def unpack(self, value):
        if six.PY3 and isinstance(value, six.binary_type):
            return super(JSONPacker, self).unpack(
                json.loads(str(value, 'utf-8')),
            )
        return super(JSONPacker, self).unpack(json.loads(value))


class JSONPackerIncludeNullFields(JSONPacker):

    _skip_none = False


packer_mapping = {
    constants.CONTENT_TYPE_APPLICATION_JSON: JSONPacker
}


def parse_content_type(value):
    # Cleanup: application/json;charset=UTF-8
    return value.split(';')[0].strip() if value else None


def get_packer(content_type):
    try:
        return packer_mapping[parse_content_type(content_type)]
    except KeyError:
        # TODO(Eugene Frolov): Specify Exception Type and message
        raise Exception("Packer can't found for content type %s " %
                        content_type)


def set_packer(content_type, packer_class):
    packer_mapping[content_type] = packer_class
